/*  picoLZO.h--  The fast data compression functions from miniLZO,
                 rewrite and rename to use in Sphinx C-- by Kai-Hong Chen.

    Notice:  These two functions can only be used in win32 console mode now.
             So, please add "#pragma option w32c" in start of program.
             (Porting to DOS should be easy, someday I will try)

    picolzo_compress(byte *src,dword src_len,byte *dst,dword *dst_len,byte *wrkmem);
    picolzo_decompress(byte *src,dword src_len,byte *dst,dword *dst_len,NULL);

    *wrkmem must be a 65536 bytes memory block

    ---------------------------------------------------------------------------------

    miniLZO -- mini subset of the LZO real-time data compression library

    Markus F.X.J. Oberhumer
    <markus@oberhumer.com>
    http://www.oberhumer.com/opensource/lzo/

*/

#define LZO_E_OK                    0
#define LZO_E_INPUT_OVERRUN         -4
#define LZO_E_INPUT_NOT_CONSUMED    -8

dword _picolzo_compress(byte *in,dword in_len,byte *out,dword *out_len,dword *dict)
{
byte *in_end;
byte *ip_end;
byte *ii;

	EAX = in_end = in + in_len;
	ip_end = EAX - 13;
	EDI = out;
	EBX = in;
	ii = EBX;
	EBX += 4;
	for(;;){
		ESI=DSBYTE[EBX+3]<<6^DSBYTE[EBX+2]<<5^DSBYTE[EBX+1]<<5^DSBYTE[EBX]*0x21>>5&0x3FFF;
		ECX=DSDWORD[ESI*4+dict];
		IF(EBX-in<=ECX)GOTO LITERAL;
		ECX=EAX=EBX-in-ECX;
		IF(EAX>0xbfff)GOTO LITERAL;
		EDX = EBX - EAX;
		IF((EAX<=0x0800)||(DSBYTE[EDX+3]==DSBYTE[EBX+3])){
			IF(DSDWORD[EDX]^DSDWORD[EBX]&0xFFFFFF==0)GOTO MATCH;
		}
LITERAL:
		DSDWORD[ESI*4+dict]=EBX-in;;
		EBX++;
		IF(EBX>=ip_end)break;
		continue;
MATCH:
		DSDWORD[ESI*4+dict]=EBX-in;
		IF(EBX-ii>0){
			$push ecx
			ECX=EBX-ii;
			IF(ECX<=3)DSBYTE[EDI-2]|=ECX;
			ELSE IF(ECX<=18){
				DSBYTE[EDI]=ECX-3;
				EDI++;
			}
			ELSE{
				$push EDX
				EDX=ECX-18;
				AL=0;
				$stosb
				WHILE(EDX>255){
					EDX-=255;
					$stosb
				}
				DSBYTE[EDI]=EDX;EDI++;
				$pop EDX
			}
			ESI=ii;
			$rep $movsb
			ii=ESI;
			$pop ecx
		}
		EBX+=3;
		IF(DSBYTE[EDX+3]-DSBYTE[EBX])goto HERE;
		EBX++;
		IF(DSBYTE[EDX+4]-DSBYTE[EBX])goto HERE;
		EBX++;
		IF(DSBYTE[EDX+5]-DSBYTE[EBX])goto HERE;
		EBX++;
		IF(DSBYTE[EDX+6]-DSBYTE[EBX])goto HERE;
		EBX++;
		IF(DSBYTE[EDX+7]-DSBYTE[EBX])goto HERE;
		EBX++;
		IF(DSBYTE[EDX+8]-DSBYTE[EBX])goto HERE;
		EBX++;
		$push edx
		EDX = EDX + 9;
		$push ecx
		ECX = in_end;
		WHILE((EBX < ECX) && (DSBYTE[EDX] == DSBYTE[EBX])){
			EDX++;
			EBX++;
		}
		$pop ecx
		EDX = EBX - ii;
		IF(ECX <= 0x4000){
			ECX--;
			IF(EDX <= 33){
				DSBYTE[EDI] = EDX-2|32; EDI++;
			}
			ELSE{
				EDX -= 33;
				DSBYTE[EDI] = 32; EDI++;
				GOTO M3_M4_LEN;
			}
		}
		ELSE{
			ECX -= 0x4000;
			DSBYTE[EDI] = dword ECX & 0x4000 >> 11 | 16;
			IF(EDX <= 9)DSBYTE[EDI] |= EDX - 2;
			ELSE{
				EDX -= 9;
				EDI++;
M3_M4_LEN:
				AL = 0;
				WHILE(EDX > 255){
					EDX -= 255;
					$stosb
				}
				DSBYTE[EDI] = EDX;
			}
			EDI++;
		}
M3_M4_OFFSET:
		$pop EDX
		DSBYTE[EDI] = ECX & 63 << 2; EDI++;
		DSBYTE[EDI] = dword ECX >> 6; EDI++;
		GOTO THERE;
HERE:
		$push EDX
		EDX=EBX-ii;
		ECX--;
		IF(ECX<0x0800){
			DSBYTE[EDI]=dword EDX-1<<5;
			EAX=ECX & 7 << 2;
			DSBYTE[EDI] |= AL;
			EDI++;
			DSBYTE[EDI] = dword ECX >> 3; EDI++;
		}
		ELSE IF(ECX < 0x4000){
			DSBYTE[EDI] = EDX - 2 | 32; EDI++;
			GOTO M3_M4_OFFSET;
		}
		ELSE{
			ECX -= 0x3FFF;
			DSBYTE[EDI] = dword ECX & 0x4000 >> 11 | 16;
			DSBYTE[EDI] |= EDX - 2;
			EDI++;
			GOTO M3_M4_OFFSET;
		}
		$pop EDX
THERE:
		ii=EBX;
		IF(EBX>=ip_end)BREAK;
	}
	DSDWORD[out_len]=EDI-out;
	EAX=in_end-ii;
	return EAX;
}

int picolzo_compress(byte *in,dword in_len,byte *out,dword *out_len,dword *wrkmem)
{
	EDI=out;
	IF(in_len<=13)ECX=in_len;
	ELSE{
		$push edi
		ECX = _picolzo_compress(in,in_len,EDI,out_len,wrkmem);
		$pop edi
		EDI += DSDWORD[out_len];
	}
	IF(ECX > 0){
		EBX = in + in_len - ECX;
		IF((EDI == out) && (ECX <= 238)){
			DSBYTE[EDI] = 17 + ECX; EDI++;
		}
		ELSE IF(ECX <= 3){
		    DSBYTE[EDI - 2] |= ECX;
		}
		ELSE IF(ECX <= 18){
			DSBYTE[EDI] = ECX - 3; EDI++;
		}
		ELSE{
			EDX = ECX - 18;
			AL = 0;
			DSBYTE[EDI] = AL; EDI++;
			WHILE(EDX > 255){
				EDX -= 255;
				$stosb
			}
			DSBYTE[EDI] = EDX; EDI++;
		}
		ESI = EBX;
		$rep $movsb
  }
	DSDWORD[EDI] = 17;
  DSDWORD[out_len]=EDI-out+3;
	return LZO_E_OK;
}

int picolzo_decompress(byte *in,dword in_len,byte *out,dword *out_len)
{
	DSDWORD[out_len]=0;
	EDI=out;
	EBX=in;
	IF(DSBYTE[EBX] > 17){
		ECX = DSBYTE[EBX] - 17; EBX++;
		IF (ECX < 4) goto MATCH_NEXT;
		ESI = EBX;
		$rep $movsb
		EBX = ESI;
		GOTO FIRST_LITERAL_RUN;
	}
	for(;;){
		ECX = DSBYTE[EBX]; EBX++;
		IF(ECX >= 16) GOTO MATCH;
		IF(ECX==0){
			WHILE(DSBYTE[EBX] == 0){
				ECX += 255;
				EBX++;
			}
			AL = DSBYTE[EBX];
			ECX += AL + 15; EBX++;
		}
		ECX+=3;
		ESI = EBX;
		$rep $movsb
		EBX = ESI;
FIRST_LITERAL_RUN:
		ECX = DSBYTE[EBX]; EBX++;
		IF(ECX >= 16) GOTO MATCH;
		EDX = EDI - 1 - 0x0800;
		EDX -= ECX >> 2;
		EDX -= DSBYTE[EBX] << 2; EBX++;
		DSDWORD[EDI] = DSDWORD[EDX]; EDI += 3; EDX += 2;
		goto MATCH_DONE;
		for(;;){
MATCH:
	    IF(ECX >= 64){
				EDX = EDI - 1;
				EDX -= ECX >> 2 & 7;
				EDX -= DSBYTE[EBX] << 3; EBX++;
				ECX = ECX >> 5 - 1;
				goto COPY_MATCH;
	    }
			IF(ECX >= 32){
				ECX &= 31;
				IF(ZEROFLAG){
					WHILE(DSBYTE[EBX] == 0){
						ECX += 255;
						EBX++;
					}
					AL = DSBYTE[EBX];
					ECX += AL + 31; EBX++;
				}
				EDX = EDI - 1;
				EDX -= DSBYTE[EBX] >> 2;
				EDX -= DSBYTE[EBX + 1] << 6;
				EBX += 2;
	    }
	    ELSE IF(ECX >= 16){
				EDX = EDI;
				EDX -= ECX & 8 << 11;
				ECX &= 7;
				IF(ZEROFLAG){
			    WHILE(DSBYTE[EBX] == 0){
						ECX += 255;
						EBX++;
			    }
					AL = DSBYTE[EBX];
			    ECX += AL + 7; EBX++;
				}
				EDX -= DSBYTE[EBX] >> 2; EBX++;
				EDX -= DSBYTE[EBX] << 6; EBX++;
				IF(EDX == EDI) BREAK 1;
				EDX -= 0x4000;
	    }
	    ELSE{
				EDX = EDI - 1;
				EDX -= ECX >> 2;
				EDX -= DSBYTE[EBX] << 2; EBX++;
				DSWORD[EDI] = DSWORD[EDX]; EDI += 2; EDX++;
				GOTO MATCH_DONE;
	    }
COPY_MATCH:
			ECX+=2;
			ESI = EDX;
			$rep $movsb
			EDX = ESI;
MATCH_DONE:
	    ECX = DSBYTE[EBX - 2] & 3;
	    IF(ZEROFLAG)BREAK;
MATCH_NEXT:
			ESI = EBX;
			$rep $movsb
			$lodsb
			EBX = ESI;
	    ECX = AL;
		}
	}
	DSDWORD[out_len] = EDI - out;
	IF(in + in_len == EBX) RETURN LZO_E_OK;
	IF(!CARRYFLAG) RETURN LZO_E_INPUT_NOT_CONSUMED;
	return LZO_E_INPUT_OVERRUN;
}
